//
// Accounts or entries into the hiscore list. 
// An account holds all data needed to rank a player, and some extra.
//
// Warning: The Account struct/class doesn't have any constructors/destructor, so all initialization and 
//          linking is done manually. So malloc()/free() is used to indicate this. Sadly can lead to harder 
//          to read code and add bugs in the future.
//

#include "Account.h"

#include "HiscoreOrder.h"
#include "ScoreValidationDec.h"

#define _CRT_SECURE_NO_WARNINGS
#include <string.h>
#include <stdlib.h>

Account *				pAccountNormalRoot	= NULL;			// Root for normal mode hiscore list
Account *				pAccountMasterRoot	= NULL;			// Root for master mode hiscore list
Account *				pAccountPlusRoot	= NULL;			// Root for plus mode hiscore list
Account *				pAccountDeathRoot	= NULL;			// Root for death mode hiscore list

// So we know how much to scroll the window
unsigned int			uNumNormalAccounts	= 0U;
unsigned int			uNumMasterAccounts	= 0U;
unsigned int			uNumPlusAccounts	= 0U;
unsigned int			uNumDeathAccounts	= 0U;

//
// Ranking system:
// The hiscore is ranked in order better or equal and newest to oldest. The way to see and test for equal hiscore is to OR the 
// signed long results of all tests together and see if it equals 0. If the result is a negative number then the test is false, 
// else if the result is 0 or a positive number then the test is true. A result of 0 means it is the exact same and the score 
// validation code should be checked to discard duplicate entries.
//

inline signed long LastHigherEqual(unsigned long dwUs, unsigned long dwOther)
{
	return (signed long)(dwUs - dwOther);
}

inline signed long LastLowerEqual(unsigned long dwUs, unsigned long dwOther)
{
	return (signed long)(dwOther - dwUs);
}

signed long RankGradeLine(Account * pAccountUs, Account * pAccountOther)
{
	return LastHigherEqual(
			((unsigned long)pAccountUs->wGrade << 2) + (unsigned long)pAccountUs->wRankedLine, 
			((unsigned long)pAccountOther->wGrade << 2) + (unsigned long)pAccountOther->wRankedLine);
}

signed long RankTime(Account * pAccountUs, Account * pAccountOther)
{
	return LastLowerEqual(
			pAccountUs->dwTime, 
			pAccountOther->dwTime);
}

signed long RankScore(Account * pAccountUs, Account * pAccountOther)
{
	return LastHigherEqual(
			pAccountUs->dwScore, 
			pAccountOther->dwScore);
}

signed long RankLevel(Account * pAccountUs, Account * pAccountOther)
{
	if (pAccountUs->bMode == LTAP_GAMEMODE_NORMAL)
	{
		// Normal Mode always ranks lower levels as higher
		return LastLowerEqual(
				(unsigned long)pAccountUs->wLevel, 
				(unsigned long)pAccountOther->wLevel);
	}

	return LastHigherEqual(
			(unsigned long)pAccountUs->wLevel, 
			(unsigned long)pAccountOther->wLevel);
}

signed long RankMedalDecoration(Account * pAccountUs, Account * pAccountOther)
{
	// Must be 0 for equal to response
	return 0L;
}

signed long RankMedalAC(Account * pAccountUs, Account * pAccountOther)
{
	return LastHigherEqual(
			(unsigned long)pAccountUs->abMedallions[INDEX_ORDER_AC], 
			(unsigned long)pAccountOther->abMedallions[INDEX_ORDER_AC]);
}

signed long RankMedalST(Account * pAccountUs, Account * pAccountOther)
{
	return LastHigherEqual(
			(unsigned long)pAccountUs->abMedallions[INDEX_ORDER_ST], 
			(unsigned long)pAccountOther->abMedallions[INDEX_ORDER_ST]);
}

signed long RankMedalSK(Account * pAccountUs, Account * pAccountOther)
{
	return LastHigherEqual(
			(unsigned long)pAccountUs->abMedallions[INDEX_ORDER_SK], 
			(unsigned long)pAccountOther->abMedallions[INDEX_ORDER_SK]);
}

signed long RankMedalRE(Account * pAccountUs, Account * pAccountOther)
{
	return LastHigherEqual(
			(unsigned long)pAccountUs->abMedallions[INDEX_ORDER_RE], 
			(unsigned long)pAccountOther->abMedallions[INDEX_ORDER_RE]);
}

signed long RankMedalRO(Account * pAccountUs, Account * pAccountOther)
{
	return LastHigherEqual(
			(unsigned long)pAccountUs->abMedallions[INDEX_ORDER_RO], 
			(unsigned long)pAccountOther->abMedallions[INDEX_ORDER_RO]);
}

signed long RankMedalCO(Account * pAccountUs, Account * pAccountOther)
{
	return LastHigherEqual(
			(unsigned long)pAccountUs->abMedallions[INDEX_ORDER_CO], 
			(unsigned long)pAccountOther->abMedallions[INDEX_ORDER_CO]);
}

signed long (* aScoreCmp[10]) (Account *, Account *) = 
{
	NULL, 
	NULL, 
	NULL, 
	NULL, 
	NULL, 
	NULL, 
	NULL, 
	NULL, 
	NULL, 
	NULL
};

#define	cpScoreCmpEnd		&aScoreCmp[10]

void BuildRankingPriority()
{
	// The array above is a list of functions used to compare a new account entry with a previously added one. The order is important and decided 
	// by the file ORDER.TXT. ORDER.TXT also holds the order of medallions for optional ranking of medallions which can spice things up a little. 
	// So there are 10 compare functions that we need to add into the list, and we go from top to bottom, and once we encounter medallions we 
	// either add non-compare functions for default TAP hiscore functionality, or we add the customized order of medallions. 
	// In the end we must have every single slot in the list filled up with functions, as there muse be no NULL pointers!
	signed long (* (* pScoreCmpCurrent)) (Account *, Account *) = &aScoreCmp[0];

	unsigned int i, j;
	unsigned char bApplyOnce = 0x0;

	for (i = 0; i < (unsigned int)sizeof (bCaptionOrder); i++)
	{
		switch (bCaptionOrder[i])
		{
		case INDEX_ORDER_MEDAL:
			if (!(bApplyOnce & 0x1))
			{
				bApplyOnce |= 0x1;

				if (FDoNotRankMedallions == true)
				{
					// Just like TAP, medallions are not worth anything other than decorations
					*pScoreCmpCurrent++ = &RankMedalDecoration;
					*pScoreCmpCurrent++ = &RankMedalDecoration;
					*pScoreCmpCurrent++ = &RankMedalDecoration;
					*pScoreCmpCurrent++ = &RankMedalDecoration;
					*pScoreCmpCurrent++ = &RankMedalDecoration;
					*pScoreCmpCurrent++ = &RankMedalDecoration;
				}
				else
				{
					for (j = 0; j < (unsigned int)sizeof (bMedallionOrder); j++)
					{
						switch (bMedallionOrder[j])
						{
						case INDEX_ORDER_AC:
							{
								*pScoreCmpCurrent++ = &RankMedalAC;
							}
							break;

						case INDEX_ORDER_ST:
							{
								*pScoreCmpCurrent++ = &RankMedalST;
							}
							break;

						case INDEX_ORDER_SK:
							{
								*pScoreCmpCurrent++ = &RankMedalSK;
							}
							break;

						case INDEX_ORDER_RE:
							{
								*pScoreCmpCurrent++ = &RankMedalRE;
							}
							break;

						case INDEX_ORDER_RO:
							{
								*pScoreCmpCurrent++ = &RankMedalRO;
							}
							break;

						case INDEX_ORDER_CO:
							{
								*pScoreCmpCurrent++ = &RankMedalCO;
							}
							break;
						}
					}
				}
			}
			break;

		case INDEX_ORDER_LEVEL:
			if (!(bApplyOnce & 0x2))
			{
				bApplyOnce |= 0x2;

				*pScoreCmpCurrent++ = &RankLevel;
			}
			break;

		case INDEX_ORDER_SCORE:
			if (!(bApplyOnce & 0x4))
			{
				bApplyOnce |= 0x4;

				*pScoreCmpCurrent++ = &RankScore;
			}
			break;

		case INDEX_ORDER_TIME:
			if (!(bApplyOnce & 0x8))
			{
				bApplyOnce |= 0x8;

				*pScoreCmpCurrent++ = &RankTime;
			}
			break;

		case INDEX_ORDER_GRADE:
			if (!(bApplyOnce & 0x10))
			{
				bApplyOnce |= 0x10;

				*pScoreCmpCurrent++ = &RankGradeLine;
			}
			break;
		}
	}

	if ((bApplyOnce & 0x1F) != 0x1F)
	{
		while (pScoreCmpCurrent != cpScoreCmpEnd)
		{
			// Missed order entries!
			*pScoreCmpCurrent++ = &RankMedalDecoration;
		}
	}
}

bool InsertAccount(Account ** ppAccountFrom, Account * pAccountNew)
{
	if (ppAccountFrom == NULL)
	{
		return false;
	}

	if (*ppAccountFrom == NULL)
	{
		pAccountNew->pNext = NULL;
		*ppAccountFrom = pAccountNew;

		return true;
	}
	else
	{
		Account * pAccountTest = (*ppAccountFrom), * pAccountPrevious = NULL;

		do
		{
			// Add the account above the account with the same ranking score, but don't add if that 
			// account uses the exact same score validation code
			signed long (* (* pScoreCmpCurrent)) (Account *, Account *) = &aScoreCmp[0];
			signed long lRanked = 0;

			while (pScoreCmpCurrent != cpScoreCmpEnd && 
				   lRanked == 0)
			{
				// lRanked will be 0 if pAccountNew and pAccountTest were equal, and positive if it should be inserted
				lRanked |= (*pScoreCmpCurrent++)(pAccountNew, pAccountTest);
			}

			if (lRanked >= 0L)
			{
				if (lRanked == 0)
				{
					// Test the score validation code to see if we can be added or not
					if (strncmp(pAccountTest->aCode, pAccountNew->aCode, 16) == 0)
					{
						// No, don't accept two of the same entries
						return false;
					}
				}

				if (pAccountPrevious == NULL)
				{
					// We need to be added at the front of the whole list
					pAccountNew->pNext = (*ppAccountFrom);
					*ppAccountFrom = pAccountNew;
				}
				else
				{
					// Take over this spot by linking the previous against us instead
					pAccountNew->pNext = pAccountPrevious->pNext;
					pAccountPrevious->pNext = pAccountNew;
				}

				return true;
			}
			
			pAccountPrevious = pAccountTest;
			pAccountTest = pAccountTest->pNext;
		}
		while (pAccountTest != NULL);

		// Add us to the end of the list
		pAccountNew->pNext = NULL;
		pAccountPrevious->pNext = pAccountNew;

		return true;
	}

	return false;
}

//
// At this point in time we have no reason to remove one account entry at a time. So we don't include any function such as DeleteAccount()
//

bool AddAccount(Account * pAccountNew)
{
	//
	// The new account sent to us must have been initialized so that 
	// the data can be tested right here
	//

	if (pAccountNew == NULL)
	{
		return false;
	}

	Account ** ppAccountRoot = NULL;
	unsigned int * puNumAccounts;

	switch (pAccountNew->bMode)
	{
	case LTAP_GAMEMODE_NORMAL:
		ppAccountRoot = &pAccountNormalRoot;
		puNumAccounts = &uNumNormalAccounts;
		break;

	case LTAP_GAMEMODE_MASTER:
		ppAccountRoot = &pAccountMasterRoot;
		puNumAccounts = &uNumMasterAccounts;
		break;

	case LTAP_GAMEMODE_PLUS:
		ppAccountRoot = &pAccountPlusRoot;
		puNumAccounts = &uNumPlusAccounts;
		break;

	case LTAP_GAMEMODE_DEATH:
		ppAccountRoot = &pAccountDeathRoot;
		puNumAccounts = &uNumDeathAccounts;
		break;
	}

	if (ppAccountRoot == NULL)
	{
		return false;
	}

	bool FReturn = InsertAccount(ppAccountRoot, pAccountNew);

	if (FReturn == true)
	{
		++(*puNumAccounts);
	}

	return FReturn;
}

void DeleteAccounts(Account ** ppAccountRoot)
{
	if (ppAccountRoot == NULL)
	{
		return;
	}

	if (*ppAccountRoot != NULL)
	{
		Account * pAccount = *ppAccountRoot;

		while (pAccount != NULL)
		{
			Account * pAccountNext = pAccount->pNext;

			// No destructor, no time to waste. Clean up manually
			if (pAccount->pszComment != NULL)
			{
				free(pAccount->pszComment);
			}

			free(pAccount);

			pAccount = pAccountNext;
		}

		*ppAccountRoot = NULL;
	}
}

void CleanupAccounts()
{
	DeleteAccounts(&pAccountNormalRoot);
	DeleteAccounts(&pAccountMasterRoot);
	DeleteAccounts(&pAccountPlusRoot);
	DeleteAccounts(&pAccountDeathRoot);

	uNumNormalAccounts	= 0U;
	uNumMasterAccounts	= 0U;
	uNumPlusAccounts	= 0U;
	uNumDeathAccounts	= 0U;
}

void AddModeHeaders()
{
	// So in order for us to scroll hiscore lists that are longer than the display area, we need a easy way to calculate the 
	// start of the scroll. Now if you think about it, dividing lists into different modes and then adding headers and empty 
	// lines don't make it easy. That's why we create spacial accounts that act as empty lines or headers so we can use basic 
	// calculations to find the start of the scroll. 
	// The new accounts need to be identified as either a mode header or an empty line, and they need to stay ordered in their 
	// lists. I think I would go for the format:
	//
	// List start:
	// EMPTY LINE
	// EMPTY LINE
	// HEADER TOP LINE
	// MODE HEADER
	// HEADER BOTTOM LINE
	// 1  First entry
	// 2  Second entry
	// 3  ...
	//
	// An empty line would have to be the highest possible rank, so we will keep define it as having maximum values
	unsigned int i, j;
	Account * pAccountNew;

	// Create the header top line for Normal Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Normal Mode
		pAccountNew->bMode = LTAP_GAMEMODE_NORMAL;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header top line is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 1);
		}

		pAccountNew->wGrade = (0x7FFF - 1);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 1);
		pAccountNew->dwTime = (0x0 + 1);
		pAccountNew->dwScore = (0x7FFFFFFF - 1);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 1);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountNormalRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumNormalAccounts;
	}
	while (false);

	// Create the header for Normal Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Normal Mode
		pAccountNew->bMode = LTAP_GAMEMODE_NORMAL;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 2);
		}

		pAccountNew->wGrade = (0x7FFF - 2);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 2);
		pAccountNew->dwTime = (0x0 + 2);
		pAccountNew->dwScore = (0x7FFFFFFF - 2);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 2);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountNormalRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumNormalAccounts;
	}
	while (false);

	// Create the header bottom line for Normal Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Normal Mode
		pAccountNew->bMode = LTAP_GAMEMODE_NORMAL;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header top line is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 3);
		}

		pAccountNew->wGrade = (0x7FFF - 3);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 3);
		pAccountNew->dwTime = (0x0 + 3);
		pAccountNew->dwScore = (0x7FFFFFFF - 3);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 3);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountNormalRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumNormalAccounts;
	}
	while (false);

	// Create two empty lines at the top of the Master Mode ranking to divide Normal Mode and Master Mode
	for (i = 0; i < 2; i++)
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Master Mode
		pAccountNew->bMode = LTAP_GAMEMODE_MASTER;

		// Clear out the name
		for (j = 0; j < 3; j++)
		{
			pAccountNew->aName[j] = 0;
		}

		// The code must be unique so it can be added twice
		for (j = 0; j < 16; j++)
		{
			pAccountNew->aCode[j] = (char)((i + 1) * (j + 1));
		}

		pAccountNew->wGrade = 0x7FFF;
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = 0x7FFF;
		pAccountNew->dwTime = 0x0;
		pAccountNew->dwScore = 0x7FFFFFFF;
		pAccountNew->abMedallions[INDEX_ORDER_AC] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_ST] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_SK] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_RE] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_RO] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_CO] = 0x7F;

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountMasterRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumMasterAccounts;
	}

	// Create the header top line for Master Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Master Mode
		pAccountNew->bMode = LTAP_GAMEMODE_MASTER;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header top line is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 1);
		}

		pAccountNew->wGrade = (0x7FFF - 1);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 1);
		pAccountNew->dwTime = (0x0 + 1);
		pAccountNew->dwScore = (0x7FFFFFFF - 1);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 1);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountMasterRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumMasterAccounts;
	}
	while (false);

	// Create the header for Master Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Master Mode
		pAccountNew->bMode = LTAP_GAMEMODE_MASTER;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 2);
		}

		pAccountNew->wGrade = (0x7FFF - 2);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 2);
		pAccountNew->dwTime = (0x0 + 2);
		pAccountNew->dwScore = (0x7FFFFFFF - 2);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 2);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountMasterRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumMasterAccounts;
	}
	while (false);

	// Create the header bottom line for Master Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Master Mode
		pAccountNew->bMode = LTAP_GAMEMODE_MASTER;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header top line is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 3);
		}

		pAccountNew->wGrade = (0x7FFF - 3);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 3);
		pAccountNew->dwTime = (0x0 + 3);
		pAccountNew->dwScore = (0x7FFFFFFF - 3);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 3);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountMasterRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumMasterAccounts;
	}
	while (false);

	// Create two empty lines at the top of the Plus Mode ranking to divide Master Mode and Plus Mode
	for (i = 0; i < 2; i++)
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Plus Mode
		pAccountNew->bMode = LTAP_GAMEMODE_PLUS;

		// Clear out the name
		for (j = 0; j < 3; j++)
		{
			pAccountNew->aName[j] = 0;
		}

		// The code must be unique so it can be added twice
		for (j = 0; j < 16; j++)
		{
			pAccountNew->aCode[j] = (char)((i + 1) * (j + 1));
		}

		pAccountNew->wGrade = 0x7FFF;
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = 0x7FFF;
		pAccountNew->dwTime = 0x0;
		pAccountNew->dwScore = 0x7FFFFFFF;
		pAccountNew->abMedallions[INDEX_ORDER_AC] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_ST] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_SK] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_RE] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_RO] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_CO] = 0x7F;

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountPlusRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumPlusAccounts;
	}

	// Create the header top line for Plus Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Plus Mode
		pAccountNew->bMode = LTAP_GAMEMODE_PLUS;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header top line is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 1);
		}

		pAccountNew->wGrade = (0x7FFF - 1);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 1);
		pAccountNew->dwTime = (0x0 + 1);
		pAccountNew->dwScore = (0x7FFFFFFF - 1);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 1);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountPlusRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumPlusAccounts;
	}
	while (false);

	// Create the header for Plus Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Plus Mode
		pAccountNew->bMode = LTAP_GAMEMODE_PLUS;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 2);
		}

		pAccountNew->wGrade = (0x7FFF - 2);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 2);
		pAccountNew->dwTime = (0x0 + 2);
		pAccountNew->dwScore = (0x7FFFFFFF - 2);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 2);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountPlusRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumPlusAccounts;
	}
	while (false);

	// Create the header bottom line for Plus Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Plus Mode
		pAccountNew->bMode = LTAP_GAMEMODE_PLUS;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header top line is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 3);
		}

		pAccountNew->wGrade = (0x7FFF - 3);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 3);
		pAccountNew->dwTime = (0x0 + 3);
		pAccountNew->dwScore = (0x7FFFFFFF - 3);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 3);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountPlusRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumPlusAccounts;
	}
	while (false);

	// Create two empty lines at the top of the Death Mode ranking to divide Plus Mode and Death Mode
	for (i = 0; i < 2; i++)
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Death Mode
		pAccountNew->bMode = LTAP_GAMEMODE_DEATH;

		// Clear out the name
		for (j = 0; j < 3; j++)
		{
			pAccountNew->aName[j] = 0;
		}

		// The code must be unique so it can be added twice
		for (j = 0; j < 16; j++)
		{
			pAccountNew->aCode[j] = (char)((i + 1) * (j + 1));
		}

		pAccountNew->wGrade = 0x7FFF;
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = 0x7FFF;
		pAccountNew->dwTime = 0x0;
		pAccountNew->dwScore = 0x7FFFFFFF;
		pAccountNew->abMedallions[INDEX_ORDER_AC] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_ST] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_SK] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_RE] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_RO] = 0x7F;
		pAccountNew->abMedallions[INDEX_ORDER_CO] = 0x7F;

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountDeathRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumDeathAccounts;
	}

	// Create the header top line for Death Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Death Mode
		pAccountNew->bMode = LTAP_GAMEMODE_DEATH;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header top line is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 1);
		}

		pAccountNew->wGrade = (0x7FFF - 1);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 1);
		pAccountNew->dwTime = (0x0 + 1);
		pAccountNew->dwScore = (0x7FFFFFFF - 1);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 1);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 1);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountDeathRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumDeathAccounts;
	}
	while (false);

	// Create the header for Death Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Death Mode
		pAccountNew->bMode = LTAP_GAMEMODE_DEATH;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 2);
		}

		pAccountNew->wGrade = (0x7FFF - 2);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 2);
		pAccountNew->dwTime = (0x0 + 2);
		pAccountNew->dwScore = (0x7FFFFFFF - 2);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 2);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 2);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountDeathRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumDeathAccounts;
	}
	while (false);

	// Create the header bottom line for Death Mode
	do
	{
		// Try to allocate the account
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Identify as belonging to Death Mode
		pAccountNew->bMode = LTAP_GAMEMODE_DEATH;

		// Clear out the name
		for (i = 0; i < 3; i++)
		{
			pAccountNew->aName[i] = 0;
		}

		// Only one header top line is allowed
		for (i = 0; i < 16; i++)
		{
			pAccountNew->aCode[i] = (char)((i + 1) * 3);
		}

		pAccountNew->wGrade = (0x7FFF - 3);
		pAccountNew->wRankedLine = 0;
		pAccountNew->wLevel = (0x7FFF - 3);
		pAccountNew->dwTime = (0x0 + 3);
		pAccountNew->dwScore = (0x7FFFFFFF - 3);
		pAccountNew->abMedallions[INDEX_ORDER_AC] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_ST] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_SK] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_RE] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_RO] = (0x7F - 3);
		pAccountNew->abMedallions[INDEX_ORDER_CO] = (0x7F - 3);

		// We MUST initialize the pointer to NULL
		pAccountNew->pszComment = NULL;

		// Now add to the list
		if (InsertAccount(&pAccountDeathRoot, pAccountNew) == false)
		{
			free(pAccountNew);

			break;
		}

		// This is what makes scroll calculations so easy
		++uNumDeathAccounts;
	}
	while (false);
}
